@@ -1,15 +1,28 @@
 import base64
 import json
+import logging
 from datetime import datetime
 from enum import Enum
 from typing import List
 from urllib.parse import urljoin
 
 import requests
-from pygrocy.utils import parse_date, parse_float, parse_int, localize_datetime
+
+from pygrocy.utils import (
+    localize_datetime,
+    parse_bool_int,
+    parse_date,
+    parse_float,
+    parse_int,
+)
+
+from .errors import GrocyError
 
 DEFAULT_PORT_NUMBER = 9192
 
+_LOGGER = logging.getLogger(__name__)
+_LOGGER.setLevel(logging.INFO)
+
 
 class ShoppingListItem(object):
     def __init__(self, parsed_json):
@@ -172,12 +185,6 @@ class ProductData(object):
             parsed_json.get("default_best_before_days", None)
         )
 
-        barcodes_raw = parsed_json.get("barcode", "")
-        if barcodes_raw is None:
-            self._barcodes = None
-        else:
-            self._barcodes = barcodes_raw.split(",")
-
     @property
     def id(self) -> int:
         return self._id
@@ -190,10 +197,6 @@ class ProductData(object):
     def name(self) -> str:
         return self._name
 
-    @property
-    def barcodes(self) -> List[str]:
-        return self._barcodes
-
 
 class ChoreData(object):
     def __init__(self, parsed_json):
@@ -203,13 +206,13 @@ class ChoreData(object):
         self.period_type = parsed_json.get("period_type")
         self.period_config = parsed_json.get("period_config")
         self.period_days = parse_int(parsed_json.get("period_days"))
-        self.track_date_only = parsed_json.get("track_date_only")
-        self.rollover = parsed_json.get("rollover")
+        self.track_date_only = parse_bool_int(parsed_json.get("track_date_only"))
+        self.rollover = parse_bool_int(parsed_json.get("rollover"))
         self.assignment_type = parsed_json.get("assignment_type")
         self.assignment_config = parsed_json.get("assignment_config")
-        self.next_execution_assigned_to_user_id = parse_int(parsed_json.get(
-            "next_execution_assigned_to_user_id"
-        ))
+        self.next_execution_assigned_to_user_id = parse_int(
+            parsed_json.get("next_execution_assigned_to_user_id")
+        )
         self.userfields = parsed_json.get("userfields")
 
 
@@ -321,22 +324,41 @@ class MissingProductResponse(object):
 
 class CurrentVolatilStockResponse(object):
     def __init__(self, parsed_json):
-        self._expiring_products = [
-            CurrentStockResponse(product)
-            for product in parsed_json.get("expiring_products")
-        ]
-        self._expired_products = [
-            CurrentStockResponse(product)
-            for product in parsed_json.get("expired_products")
-        ]
-        self._missing_products = [
-            MissingProductResponse(product)
-            for product in parsed_json.get("missing_products")
-        ]
+        self._due_products = []
+        if "due_products" in parsed_json:
+            self._due_products = [
+                CurrentStockResponse(product)
+                for product in parsed_json.get("due_products")
+            ]
+
+        self._overdue_products = []
+        if "overdue_products" in parsed_json:
+            self._overdue_products = [
+                CurrentStockResponse(product)
+                for product in parsed_json.get("overdue_products")
+            ]
+
+        self._expired_products = []
+        if "expired_products" in parsed_json:
+            self._expired_products = [
+                CurrentStockResponse(product)
+                for product in parsed_json.get("expired_products")
+            ]
+
+        self._missing_products = []
+        if "missing_products" in parsed_json:
+            self._missing_products = [
+                MissingProductResponse(product)
+                for product in parsed_json.get("missing_products")
+            ]
 
     @property
-    def expiring_products(self) -> List[CurrentStockResponse]:
-        return self._expiring_products
+    def due_products(self) -> List[CurrentStockResponse]:
+        return self._due_products
+
+    @property
+    def overdue_products(self) -> List[CurrentStockResponse]:
+        return self._overdue_products
 
     @property
     def expired_products(self) -> List[CurrentStockResponse]:
@@ -347,6 +369,15 @@ class CurrentVolatilStockResponse(object):
         return self._missing_products
 
 
+class ProductBarcodeData(object):
+    def __init__(self, parsed_json):
+        self._barcode = str(parsed_json.get("barcode"))
+
+    @property
+    def barcode(self) -> str:
+        return self._barcode
+
+
 class ProductDetailsResponse(object):
     def __init__(self, parsed_json):
         self._last_purchased = parse_date(parsed_json.get("last_purchased"))
@@ -360,13 +391,20 @@ class ProductDetailsResponse(object):
 
         self._product = ProductData(parsed_json.get("product"))
 
-        self._quantity_unit_purchase = QuantityUnitData(
-            parsed_json.get("quantity_unit_purchase")
-        )
         self._quantity_unit_stock = QuantityUnitData(
             parsed_json.get("quantity_unit_stock")
         )
 
+        self._parse_location(parsed_json)
+
+        self._parse_barcodes(parsed_json)
+
+    def _parse_barcodes(self, parsed_json):
+        barcodes_raw = parsed_json.get("product_barcodes", "")
+        if barcodes_raw is not None:
+            self._barcodes = [ProductBarcodeData(barcode) for barcode in barcodes_raw]
+
+    def _parse_location(self, parsed_json):
         raw_location = parsed_json.get("location")
         if raw_location is None:
             self._location = None
@@ -397,6 +435,10 @@ class ProductDetailsResponse(object):
     def last_price(self) -> float:
         return self._last_price
 
+    @property
+    def barcodes(self) -> List[ProductBarcodeData]:
+        return self._barcodes
+
     @property
     def product(self) -> ProductData:
         return self._product
@@ -467,11 +509,55 @@ class TaskResponse(object):
         self.userfields = parsed_json.get("userfields")
 
 
+class CurrentBatteryResponse(object):
+    def __init__(self, parsed_json):
+        self.id = parse_int(parsed_json.get("battery_id"))
+        self.last_tracked_time = parse_date(parsed_json.get("last_tracked_time"))
+        self.next_estimated_charge_time = parse_date(
+            parsed_json.get("'next_estimated_charge_time")
+        )
+
+
+class BatteryData(object):
+    def __init__(self, parsed_json):
+        self.id = parse_int(parsed_json.get("id"))
+        self.name = parsed_json.get("name")
+        self.description = parsed_json.get("description")
+        self.used_in = parsed_json.get("used_in")
+        self.charge_interval_days = parse_int(parsed_json.get("charge_interval_days"))
+        self.created_timestamp = parse_date(parsed_json.get("row_created_timestamp"))
+        self.userfields = parsed_json.get("userfields")
+
+
+class BatteryDetailsResponse(object):
+    def __init__(self, parsed_json):
+        self.battery = BatteryData(parsed_json.get("battery"))
+        self.charge_cycles_count = parse_int(parsed_json.get("charge_cycles_count"))
+        self.last_charged = parse_date(parsed_json.get("last_charged"))
+        self.next_estimated_charge_time = parse_date(
+            parsed_json.get("'next_estimated_charge_time")
+        )
+
+
+def _enable_debug_mode():
+    _LOGGER.setLevel(logging.DEBUG)
+
+
 class GrocyApiClient(object):
     def __init__(
-            self, base_url, api_key, port: int = DEFAULT_PORT_NUMBER, verify_ssl=True
+        self,
+        base_url,
+        api_key,
+        port: int = DEFAULT_PORT_NUMBER,
+        verify_ssl=True,
+        debug=False,
     ):
+        if debug:
+            _enable_debug_mode()
+
         self._base_url = "{}:{}/api/".format(base_url, port)
+        _LOGGER.debug(f"generated base url: {self._base_url}")
+
         self._api_key = api_key
         self._verify_ssl = verify_ssl
         if self._api_key == "demo_mode":
@@ -481,25 +567,35 @@ class GrocyApiClient(object):
 
     def _do_get_request(self, end_url, params=None):
         req_url = urljoin(self._base_url, end_url)
-        resp = requests.get(req_url,
-                            params=params,
-                            verify=self._verify_ssl,
-                            headers=self._headers)
-        resp.raise_for_status()
+        resp = requests.get(
+            req_url,
+            params=params,
+            verify=self._verify_ssl,
+            headers=self._headers)
+
+        _LOGGER.debug("-->\tGET /%s", end_url)
+        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
+        _LOGGER.debug("\t\t%s", resp.content)
+
+        if resp.status_code >= 400:
+            raise GrocyError(resp)
+
         if len(resp.content) > 0:
             return resp.json()
 
-    def _do_delete_request(self, end_url: str):
-        req_url = urljoin(self._base_url, end_url)
-        resp = requests.delete(req_url, verify=self._verify_ssl, headers=self._headers)
-        resp.raise_for_status()
-
     def _do_post_request(self, end_url: str, data: dict):
         req_url = urljoin(self._base_url, end_url)
         resp = requests.post(
             req_url, verify=self._verify_ssl, headers=self._headers, json=data
         )
-        resp.raise_for_status()
+
+        _LOGGER.debug("-->\tPOST /%s", end_url)
+        _LOGGER.debug("\t\t%s", data)
+        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
+        _LOGGER.debug("\t\t%s", resp.content)
+
+        if resp.status_code >= 400:
+            raise GrocyError(resp)
         if len(resp.content) > 0:
             return resp.json()
 
@@ -515,7 +611,29 @@ class GrocyApiClient(object):
         resp = requests.put(
             req_url, verify=self._verify_ssl, headers=up_header, data=data
         )
-        resp.raise_for_status()
+
+        _LOGGER.debug("-->\tPUT /%s", end_url)
+        _LOGGER.debug("\t\t%s", data)
+        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
+        _LOGGER.debug("\t\t%s", resp.content)
+
+        if resp.status_code >= 400:
+            raise GrocyError(resp)
+
+        if len(resp.content) > 0:
+            return resp.json()
+
+    def _do_delete_request(self, end_url: str):
+        req_url = urljoin(self._base_url, end_url)
+        resp = requests.get(req_url, verify=self._verify_ssl, headers=self._headers)
+
+        _LOGGER.debug("-->\tDELETE /%s", end_url)
+        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
+        _LOGGER.debug("\t\t%s", resp.content)
+
+        if resp.status_code >= 400:
+            raise GrocyError(resp)
+
         if len(resp.content) > 0:
             return resp.json()
 
@@ -544,10 +662,10 @@ class GrocyApiClient(object):
             return ChoreDetailsResponse(parsed_json)
 
     def execute_chore(
-            self,
-            chore_id: int,
-            done_by: int = None,
-            tracked_time: datetime = datetime.now(),
+        self,
+        chore_id: int,
+        done_by: int = None,
+        tracked_time: datetime = datetime.now(),
     ):
         localized_tracked_time = localize_datetime(tracked_time)
 
@@ -559,12 +677,12 @@ class GrocyApiClient(object):
         return self._do_post_request(f"chores/{chore_id}/execute", data)
 
     def add_product(
-            self,
-            product_id,
-            amount: float,
-            price: float,
-            best_before_date: datetime = None,
-            transaction_type: TransactionType = TransactionType.PURCHASE,
+        self,
+        product_id,
+        amount: float,
+        price: float,
+        best_before_date: datetime = None,
+        transaction_type: TransactionType = TransactionType.PURCHASE,
     ):
         data = {
             "amount": amount,
@@ -578,11 +696,11 @@ class GrocyApiClient(object):
         return self._do_post_request(f"stock/products/{product_id}/add", data)
 
     def consume_product(
-            self,
-            product_id: int,
-            amount: float = 1,
-            spoiled: bool = False,
-            transaction_type: TransactionType = TransactionType.CONSUME,
+        self,
+        product_id: int,
+        amount: float = 1,
+        spoiled: bool = False,
+        transaction_type: TransactionType = TransactionType.CONSUME,
     ):
         data = {
             "amount": amount,
@@ -604,7 +722,7 @@ class GrocyApiClient(object):
         self._do_post_request("stock/shoppinglist/add-missing-products", data)
 
     def add_product_to_shopping_list(
-            self, product_id: int, shopping_list_id: int = 1, amount: int = 1
+        self, product_id: int, shopping_list_id: int = 1, amount: int = 1
     ):
         data = {
             "product_id": product_id,
@@ -619,7 +737,7 @@ class GrocyApiClient(object):
         self._do_post_request("stock/shoppinglist/clear", data)
 
     def remove_product_in_shopping_list(
-            self, product_id: int, shopping_list_id: int = 1, amount: int = 1
+        self, product_id: int, shopping_list_id: int = 1, amount: int = 1
     ):
         data = {
             "product_id": product_id,
@@ -682,9 +800,30 @@ class GrocyApiClient(object):
         if parsed_json:
             return RecipeDetailsResponse(parsed_json)
 
-    def add_generic(self, entity_type: str, data: dict):
-        response = self._do_post_request(f"objects/{entity_type}", data)
-        return response.get("created_object_id")
+    def get_batteries(self) -> List[CurrentBatteryResponse]:
+        parsed_json = self._do_get_request(f"batteries")
+        if parsed_json:
+            return [CurrentBatteryResponse(data) for data in parsed_json]
 
-    def delete_entity(self, entity_type: str, id):
-        self._do_delete_request(f"objects/{entity_type}/{id}")
+    def get_battery(self, battery_id: int) -> BatteryDetailsResponse:
+        parsed_json = self._do_get_request(f"batteries/{battery_id}")
+        if parsed_json:
+            return BatteryDetailsResponse(parsed_json)
+
+    def charge_battery(self, battery_id: int, tracked_time: datetime = datetime.now()):
+        localized_tracked_time = localize_datetime(tracked_time)
+        data = {"tracked_time": localized_tracked_time.isoformat()}
+
+        return self._do_post_request(f"batteries/{battery_id}/charge", data)
+
+    def add_generic(self, entity_type: str, data):
+        return self._do_post_request(f"objects/{entity_type}", data)
+
+    def update_generic(self, entity_type: str, object_id: int, data):
+        return self._do_put_request(f"objects/{entity_type}/{object_id}", data)
+
+    def delete_generic(self, entity_type: str, object_id: int):
+        return self._do_delete_request(f"objects/{entity_type}/{object_id}")
+
+    def get_generic_objects_for_type(self, entity_type: str):
+        return self._do_get_request(f"objects/{entity_type}")
